'use strict';

const eDeviceFamily = {
    MSP430:430,
    MSP432:432
};
const eMode = {
    P_Analog:        {id:0, name:"EnergyTrace"  }, 
    P_Digital:       {id:1, name:"EnergyTrace??"}, 
    P_AnalogDigital: {id:2, name:"EnergyTrace++"}
};
const eHDR = {
    Lo : {id:0},
    Hi : {id:1}
};
const eJStateFormat = {
    Unknown:   {id:0},
    Reserved1: {id:1},
    Reserved2: {id:2},
    MSP432:    {id:3},
    RF432:     {id:4}
};
const eDriverCallback = {
    Continuous:      {id:0},
    DuringRun:       {id:1},
    DuringHWTrigger: {id:2}
};

class EnergyTraceProcessor {

    constructor(core) {
        this.core = core;
        // driver paramter
        this._resetConfig();
        // stream
        this.streamID = undefined;
        this.recs = [];
        this.poll = 1000;   // 1000ms poll
        // status
        this.status = { };
    }
    /*
     * Get the configuration block used for stream creation
     */ 
    get configBlock() {
        let buffer = new ArrayBuffer(20);
        let cblk = new Int32Array(buffer);
        cblk[0] = this.param.mode.id;
        cblk[1] = 0;    // undefined
        cblk[2] = this.param.hdr.id;
        cblk[3] = this.param.jstate.id;
        cblk[4] = this.param.callback.id;
        // return a byte array
        let byteBuffer = new Uint8Array(buffer); 
        let r= [];
        for(let i=0; i<byteBuffer.byteLength; i++) {
            r.push(byteBuffer[i]);
        }
        return r;
    }
    _resetConfig() {
        this.param = {};
        this.param.mode = eMode.P_Analog;
        this.param.hdr = eHDR.Lo;
        this.param.jstate = eJStateFormat.Reserved2;
        this.param.callback = eDriverCallback.Continuous;
    }
    setConfig(mode, hdr, jstate, cb) {
        this._resetConfig();
        if(mode) {
            this.param.mode = mode;
        }
        if(hdr) {
            this.param.hdr = hdr;
        }
        if(jstate) {
            this.param.jstate = jstate;
        }
        if(cb) {
            this.param.callback = cb;
        }
    }
    _openStream() {
        var _self = this;
        this.streamID = undefined;
        this.recs = [];
        this._streamReceived = 0;
        this._streamElapsed = 0;
        let _cfg = this.configBlock;
        //
        if(this.param.mode == eMode.P_Analog) {
            this.recSize = 18;
        }
        else if(this.param.mode == eMode.P_AnalogDigital) {
            this.recSize = 26;
        }
        else {
            console.error('EnergyTrace: Invalid capture mode '+this.param.mode.name);
            this.recSize = 0;
            return;
        }
        //
        this.core.dataStreams.createGTIInputStream(_cfg).then(function(response) {
            _self.streamID = response.id;
            console.log('EnergyTrace stream created with id='+_self.streamID.toString(16));
            _self.status.runState = 'running';
            setTimeout(function(){
                _self._processData();    
            },_self.poll);
        }).catch(function(err) {
            console.error('EnergyTrace stream create error: '+err.message);
        });
    }
    _closeStream() {
        this.status.runState = 'halted';

        if(this.streamID == null) {
            console.error('No opened stream');
            return;
        }
        var _self = this;
        this.core.dataStreams.close(this.streamID).then(function(response) {
            console.log('EnergyTrace stream closed');
            _self.streamID = undefined;
        }).catch(function(err){
            console.error('EnergyTrace stream (' + _self.streamID.toString(16) + ') error: ' + err.message);
            _self.streamID = undefined;
        }).finally(function(){
            if(_self._captureDoneCB) {
                _self._captureDoneCB.call();
            }
        });
    }
    _processData() {
        var _self = this;
        if(this.streamID == null) {
            console.error('No opened stream');
            return;
        }
        return this.core.dataStreams.read(this.streamID, 32760).then(function(response) {
            //console.log(response.data.length + '/' + _self._streamReceived);
            _self._streamReceived += response.data.length;

            if(response.data.length === 0) {
                _self._streamElapsed += _self.poll;
                setTimeout(function(){
                    return _self._processData();    
                },_self.poll);
                return;                              
            }

            // ..... realtime processing
            let newRecs = _self._readRecords(response.data, false);
            _self.recs = _self.recs.concat(newRecs);
            let refRec = newRecs[newRecs.length-1];
            // console.log(refRec.time + ', ' + refRec.energy + ' / ' + _self._streamElapsed + ' % ' + _self.recs.length);
            
            // update status
            for(let i=0; i<newRecs.length; i++) {
                _self._postProcessSingleRec(newRecs[i]);
            }

            // check if enough data has been collected
            if(refRec.time >= _self.capPeriod) {
                _self._closeStream();
                return;
            }

            // Give some time to accumulate data before processing
            let _poll = 200;  // may want to use different polling rate here
            _self._streamElapsed += _poll;
            setTimeout(function(){
                return _self._processData();    
            },_poll);
        }).catch(function(err) {
            console.error('Read error: '+err.message);
            _self._closeStream();
        });
    }
    _postProcessSingleRec(rec) {
        // update status
        this.status.numRec++;
        this.status.curTime = rec.time;   // us
        this.status.curEnergy = rec.energy;    // nJ
        this.status.curCurrent = rec.current;  // nA
        this.status.curVoltage = rec.voltage;  // mV
        this.status.curPower = rec.power; // mW
        if(this.status.numRec === 1) {
            // first block
            this.status.minCurrent = this.status.maxCurrent = this.status.curCurrent;
            this.status.minPower = this.status.maxPower = this.status.curPower;
        }
        else {
            this.status.minCurrent = Math.min(this.status.minCurrent, this.status.curCurrent);
            this.status.maxCurrent = Math.max(this.status.maxCurrent, this.status.curCurrent);
            this.status.minPower = Math.min(this.status.minPower, this.status.curPower);
            this.status.maxPower = Math.max(this.status.maxPower, this.status.curPower);
        }
        this.status._totalCurrent += this.status.curCurrent;
        this.status.aveCurrent = this.status._totalCurrent/this.status.numRec;
        this.status._totalPower += this.status.curPower;
        this.status.avePower = this.status._totalPower/this.status.numRec;
    }
    _readRecords(blk, firstRecOnly) {
        if( (blk.length % this.recSize) != 0) {
            console.error('EnergyTrace: Invalide data block size '+blk.length);
        }

        let recs = [];
        let readOffset = 0;
        while(readOffset < blk.length) {
            let rec = this._readRecord(blk, readOffset);
            // Check invalid records
            readOffset += this.recSize;
            /* TODO: need a way to identify invalid records.
            if(rec.current === 0) {
     //bc:hack           continue;
            }
            */
            // add
            recs.push(rec);
            if(firstRecOnly) {
                break;
            }
        }
        return recs;
    }
    _readRecord(blk, offset) {
        let rec = {};
        /*  
         *             Raw:     Report
         *   Time:     uS       uS
         *   Current:  nA       nA
         *   Voltage:  mV       mV
         *   Energy:   0.1uJ    nJ
         */
        rec.time = this._readTimestamp(blk, offset);
        if(this.param.mode === eMode.P_Analog) {
            rec.current = this._readInt(blk, offset+8,  4);
            rec.voltage = this._readInt(blk, offset+12, 2);
            rec.energy  = this._readInt(blk, offset+14, 4);
        }
        else if(this.param.mode === eMode.P_AnalogDigital) {
            rec.dsrLo   = dataView._readInt(blk, offset+8,  4);
            rec.dsrHi   = dataView._readInt(blk, offset+12, 4);
            rec.current = dataView._readInt(blk, offset+16, 4);
            rec.voltage = dataView._readInt(blk, offset+20, 2);
            rec.energy  = dataView._readInt(blk, offset+22, 4);
        }
        rec.power = rec.current * rec.voltage / 1000000000.0; // mW
        // conversion
        rec.energy *= 100;  // 0.1uJ to nJ

        return rec;
    }
    _readInt(blk, offset, numBytes) {
        let result = 0;
        for(let i=0, shift=0; i<numBytes; i++, shift+=8) {
            result |= (blk[i+offset] & 0xFF) << shift;
        }
        return result;
    }
    _readTimestamp(blk, offset) {
        let lo4 = this._readInt(blk, offset+1, 4);
        let hi3 = this._readInt(blk, offset+5, 3);
        let timestamp = hi3*4294967296 + lo4;   // JS doen't have 64bits long, only 64 bits float
        return timestamp;
    }
    /*
     *  status = {
     *     // Current status
     *     duration
     *     curTime
     *     curEnergy
     *     curCurrent
     *     curVoltage
     *     curPower
     *     // Statistic
     *     aveCurrent
     *     minCurrent
     *     maxCurrent
     *     avePower
     *     minPower
     *     maxPower
     *     // Housekeeping
     *     numRec
     *     _totalCurrent
     *     _totalPower
     *     runState
     *  }
     */ 
    _resetStatus() {
        this.status.duration = this.capPeriod/1000;  // ms
        this.status.curTime = 0;
        this.status.curEnergy = 0;
        this.status.curCurrent = 0;
        this.status.curVoltage = 0;
        this.status.curPower   = 0;
        //
        this.status.aveCurrent = 0;
        this.status.minCurrent = undefined;
        this.status.maxCurrent = undefined;
        //
        this.status.avePower = 0;
        this.status.minPower = undefined;
        this.status.maxPower = undefined;
        //
        this.status.numRec = 0;
        this.status._totalCurrent = 0;
        this.status._totalPower = 0;
        this.status.runState = 'idle';
    }
    capture(period, callback) {
        this.capPeriod = period * 1000;  // ms to us
        if(callback) {
            this._captureDoneCB = callback;
        }
        else {
            delete this._captureDoneCB;
        }
        this.status.duration = period;
        this._resetStatus();
        this._openStream();
    }
    /*
     * 
     * @param fields: Array of fields to be extracted
     * @param merge:  Flag for merging specified fields into a records (optional)
     * @param max:    Maximum number of records to report (optional)
     * @param reamsple: Apply if max < total records.  (optional)
     *                  If set then will resample, else just report the first 'max' records
     * Examples:
     *    getDataArray(['time','power'])       => [[time][power]]
     *    getDataArray(['time','power'], true) => [[time,power]]
     *    getDataArray(['power'])              => [[power]]
     *    getDataArray(['power'], true)        => [power]
     */
    getDataArray(fields, merge, max, resample) {
        let scale = 1;
        let end = this.recs.length;
        let result = [];
        if(max && max > 0 && max < this.recs.length) {
            if(resample) {
                scale = this.recs.length/max;
            }
            end = max;
        }
        let idx = 0;
        if(merge) {
            for(let i=0; i<end; i++) {
                idx = i*scale;
                let rec = [];
                if(fields.length === 1) {
                    result.push(this.recs[idx][fields[0]]);
                }
                else {
                    for(let j=0; j<fields.length; j++) {
                        rec.push(this.recs[idx][fields[j]]);
                    }
                    result.push(rec);
                }
            }
        }
        else {
            for(let j=0; j<fields.length; j++) {
                result.push([]);
            }
            for(let i=0; i<end; i++) {
                idx = i*scale;
                for(let j=0; j<fields.length; j++) {
                    result[j].push(this.recs[idx][fields[j]]);
                }
            }
        }
        return result;
    }
    /**
     * Flush all data storages.
     */
    flush() {
        if(this.streamID != null) {
            _closeStream();
        }
        this.recs = [];
    }
}